from logging import Logger
from retrying import retry

from azure.mgmt.resource.managedapplications.models import ProvisioningState

from sidecar.apps_configuration_end_tracker import AppsConfigurationEndTracker
from sidecar.azure_clp.azure_clients import AzureClientsManager
from sidecar.azure_clp.azure_status_maintainer import AzureStatusMaintainer
from sidecar.azure_clp.retrying_helpers import retry_if_connection_error
from sidecar.const import DateTimeProvider
from sidecar.health_check.app_health_check_state import AppHealthCheckState
from sidecar.sandbox_start_time_updater import ISandboxStartTimeUpdater
from sidecar.utils import Utils


class AzureSandboxStartTimeUpdater(ISandboxStartTimeUpdater):
    def __init__(self, date_time_provider: DateTimeProvider, logger: Logger,
                 apps_configuration_end_tracker: AppsConfigurationEndTracker,
                 status_maintainer: AzureStatusMaintainer,
                 app_health_check_state: AppHealthCheckState,
                 sandbox_id: str,
                 clients_manager: AzureClientsManager):
        super(AzureSandboxStartTimeUpdater, self).__init__(
            app_health_check_state=app_health_check_state,
            date_time_provider=date_time_provider,
            logger=logger,
            apps_configuration_end_tracker=apps_configuration_end_tracker)
        self._clients_manager = clients_manager
        self._sandbox_id = sandbox_id
        self._status_maintainer = status_maintainer

    def _on_deployment_complete(self):
        self._logger.info('waiting for deployment to succeed')
        
        Utils.retry_on_exception(func=lambda: self._throw_if_not_succeeded(),
                                 timeout_in_sec=20*1000, interval_in_sec=15,
                                 logger=self._logger, log_every_n_attempts=4,
                                 logger_msg="waiting for deployment to succeed")

        self._logger.info('deployment succeeded!')

        self._status_maintainer.update_sandbox_start_status(self._date_time_provider.get_current_time_utc())

    @retry(stop_max_attempt_number=5, wait_fixed=1000, retry_on_exception=retry_if_connection_error)
    def _throw_if_not_succeeded(self):
        deployments = list(
            self._clients_manager.resource_client.deployments.list_by_resource_group(
                resource_group_name=self._sandbox_id,
                timeout=20))
        deployment = next((x for x in deployments if x.name.startswith('qs_')), None)

        if not deployment or deployment.properties.provisioning_state != ProvisioningState.succeeded.value:
            raise Exception('not succeeded yet')



import threading
from logging import Logger
from typing import List

from blinker import signal

from sidecar.app_instance_config_status_event_reporter import AppInstanceConfigStatusEventReporter
from sidecar.app_instance_identifier import AppInstanceIdentifier
from sidecar.app_instance_service import IAppInstanceService
from sidecar.apps_configuration_end_tracker import AppsConfigurationEndTracker
from sidecar.const import Signals, AppInstanceConfigStatus
from sidecar.model.objects import SidecarApplication
from sidecar.sandbox_deployment_state_tracker import SandboxDeploymentStateTracker
from sidecar.sandbox_error import SandboxError


class AppStatusMaintainer:
    def __init__(self, logger: Logger,
                 app_instance_service: IAppInstanceService,
                 apps_configuration_end_tracker: AppsConfigurationEndTracker,
                 sandbox_deployment_state_tracker: SandboxDeploymentStateTracker,
                 app_instance_status_event_reporter: AppInstanceConfigStatusEventReporter,
                 apps: List[SidecarApplication]):
        self._logger = logger
        self._app_instance_service = app_instance_service
        self._apps_configuration_end_tracker = apps_configuration_end_tracker
        self._sandbox_deployment_state_tracker = sandbox_deployment_state_tracker
        self._app_instance_config_status_event_reporter = app_instance_status_event_reporter
        self._apps = apps
        self._lock = threading.RLock()

    def update_status(self, app_instance_identifier: AppInstanceIdentifier, status: str):
        # we want to report the event even if the instance is stale, so doing it before update_status_if_not_stale
        # the event handling should be able to handle concurrency so calling it outside the lock
        self._safely_report_app_instance_event(app_instance_identifier, status)

        with self._lock:
            # there are some scenarios where we might attempt to set a "stale" status that will override the most
            # recent one for the app instance.
            # for example, when there is a race condition involving health-checks:
            # 1. health check started -> app instance is restarted -> new instance starts another health check ->
            #    both health checks try to update their result status, in some order
            # 2. health check started -> app instance is restarted -> new instance will report end status via api and
            #    health check will try to update its result status, in some order
            #    [this scenario is no longer relevant since the api method that allows to set end status was removed]

            # _app_instance_service.update_status_if_not_stale is expected to prevent the wrong status from being
            # updated and throw an exception in this case so the flow doesn't continue.
            # it should implement the logic that if the updating app instance (identified by app_instance_identifier)
            # is no longer a part of the sandbox, then the update should not be performed (since the result is for
            # the instance before the restart). IMPORTANT: the identifier should be such that it changes on restart.
            # the validation could've been implemented out here, but it is better performance-wise to do it inside
            # _app_instance_service.update_status_if_not_stale (the instance is being fetched anyway).

            # currently, we assume that it is not possible that the same app instance (i.e. no restart, same id)
            # will try to set an end status more than once and so if the updating app instance is not stale we just
            # set the status without checking if this instance has already set an end status and whether it should
            # be overridden.
            # we might need to handle such scenarios in the future, for example if an app instance starts a health check
            # and then reports some end status via the api -> when the health check will end, it will try to override
            # the api's status with its own result.
            self._logger.info("updating app status '{APP_ID}' to '{STATUS}'"
                              .format(APP_ID=app_instance_identifier, STATUS=status))

            self._app_instance_service.update_status_if_not_stale(app_instance_identifier, status=status)
            if status == AppInstanceConfigStatus.ERROR:
                self._app_instance_service.add_error(
                    app_instance_identifier=app_instance_identifier,
                    error=SandboxError.PrivateHealthcheckFailed(app_name=app_instance_identifier.name))

            self._apps_configuration_end_tracker.update_app_instance_config_status(
                app_instance_identifier=app_instance_identifier,
                app_instance_config_status=status)

            self._sandbox_deployment_state_tracker.on_app_instance_configuration_status_updated(
                app_instance_identifier=app_instance_identifier,
                app_instance_config_status=status)

        signal(Signals.ON_INSTANCE_UPDATE_STATUS)\
            .send(self, identifier=app_instance_identifier)

    def _safely_report_app_instance_event(self, app_instance_identifier: AppInstanceIdentifier, status: str):
        # don't want unhandled errors in event handling to ruin the rest of the flow
        try:
            self._app_instance_config_status_event_reporter.on_app_instance_configuration_status_updated(
                app_instance_identifier=app_instance_identifier,
                app_instance_config_status=status)
        except Exception as exc:
            self._logger.exception("Failed to _safely_report_app_instance_event for app '{0}' exception {1}".format(app_instance_identifier.name, str(exc)))



import threading

from sidecar.apps_configuration_end_tracker import AppConfigurationEndStatus
from sidecar.health_check.app_health_check_state import AppHealthCheckState
from sidecar.messaging_service import MessagingService
from sidecar.model.objects import EnvironmentType
from sidecar.sandbox_deployment_state_tracker import SandboxDeploymentStateTracker
from logging import Logger



class SandboxEndDeploymentNotifier:
    def __init__(self,
                 deployment_state_tracker: SandboxDeploymentStateTracker,
                 messaging_service: MessagingService,
                 app_health_state_checker: AppHealthCheckState,
                 space_id: str,
                 sandbox_id: str,
                 production_id: str,
                 env_type: EnvironmentType,
                 logger: Logger):
        self._sandbox_id = sandbox_id
        self._production_id = production_id
        self._space_id = space_id
        self._app_health_state_checker = app_health_state_checker
        self._messaging_service = messaging_service
        self._deployment_state_tracker = deployment_state_tracker
        self._is_message_sent = False
        self._lock = threading.RLock()
        self._env_type = env_type
        self._logger = logger

    # please never use git history on this, it's not my fault.
    def notify_end_deployment(self):

        with self._lock:
            if self._is_message_sent:
                return

            end_status = self._deployment_state_tracker.get_deployment_end_status()

            # In case of sandboxes we want to notify (send mail) on every deployed (even when in ActiveWithError)
            if self._env_type == EnvironmentType.Sandbox:
                message_type = 'CSMSCommon.Model.Events:SandboxDeployed'
                message_data = {
                    'SpaceId': self._space_id,
                    'SandboxId': self._sandbox_id,
                    'Status': end_status
                }
                self._messaging_service.publish(message_type, message_data)
                self._is_message_sent = True

            # In case of production we want to notify (send mail) only on complete_with_success (see story 2636)
            else:
                if self._deployment_state_tracker.all_apps_deployment_ended_with_status(AppConfigurationEndStatus.COMPLETED) \
                        and self._app_health_state_checker.all_complete_with_success():
                    message_type = 'CSMSCommon.Model.Events:ProductionDeployed'
                    message_data = {
                        'SpaceId': self._space_id,
                        'SandboxId': self._sandbox_id,
                        'ProductionId': self._production_id,
                        'Status': end_status,
                        'EnvironmentType': self._get_production_env_type()
                    }
                    self._messaging_service.publish(message_type, message_data)
                    self._is_message_sent = True

    def _get_production_env_type(self):
        if self._env_type == EnvironmentType.ProductionBlue:
            return 'blue'
        elif self._env_type == EnvironmentType.ProductionGreen:
            return 'green'
        else:
            raise ValueError("EnvironmentType value of '{}' is neither 'blue' nor 'green'".format(self._env_type))



from http import HTTPStatus


class NotReadyYetError(Exception):
    def __init__(self,
                 message: str = 'NOT_READY',
                 status: HTTPStatus = HTTPStatus.CONFLICT):
        self.message = message
        self.status = status



from sidecar.aws_status_maintainer import AWSStatusMaintainer
from sidecar.sandbox_terminating_query import ISandboxTerminatingQuery


class AwsSandboxTerminatingQuery(ISandboxTerminatingQuery):

    def __init__(self, status_maintainer: AWSStatusMaintainer):
        super().__init__()
        self._status_maintainer = status_maintainer

    def is_terminating(self) -> bool:
        return self._status_maintainer.get_terminating_flag()



from marshmallow import Schema, fields, post_load, validates_schema, ValidationError

from sidecar.const import ServiceType
from sidecar.messaging_service import MessagingConnectionProperties
from sidecar.model.objects import AwsSidecarConfiguration, AzureSidecarConfiguration, \
    KubernetesSidecarConfiguration, SidecarApplication, SidecarTerraformService, Script, \
    TerraformServiceModule, InputParameter


class MessagingConnectionPropertiesSchema(Schema):
    address = fields.Str()
    exchange = fields.Str()

    @post_load
    def make(self, data):
        return MessagingConnectionProperties(**data)

    @validates_schema
    def validate_numbers(self, data):
        if not data:
            raise ValidationError("Missing MessagingConnectionProperties")

    class Meta:
        strict = True


class ScriptSchema(Schema):
    path = fields.Str(required=True)
    name = fields.Str(required=True)
    headers = fields.Dict(required=True)

    @post_load
    def make(self, data):
        return Script(**data)

    class Meta:
        strict = True


class TerraformServiceModuleSchema(Schema):
    source = fields.Str(required=False, allow_none=True)
    version = fields.Str(required=False, allow_none=True)

    @post_load
    def make(self, data):
        return TerraformServiceModule(**data)

    class Meta:
        strict = True


class InputParameterSchema(Schema):
    name = fields.Str(required=True)
    value = fields.Str(required=True)

    @post_load
    def make(self, data):
        return InputParameter(**data)

    class Meta:
        strict = True


class SidecarServiceSchema(Schema):
    name = fields.Str(required=True)
    type = fields.Str(required=True)
    dependencies = fields.List(fields.Str)
    inputs = fields.List(fields.Nested(InputParameterSchema()), allow_none=True)
    outputs = fields.List(fields.Str, allow_none=True, required=False)

    # TODO: find a better way to deal with it in the future ...
    terraform_module = fields.Nested(TerraformServiceModuleSchema(), required=False, allow_none=True)
    terraform_version = fields.Str(required=False, allow_none=True)
    tfvars_file = fields.Nested(ScriptSchema(), required=False, allow_none=True)

    @post_load
    def make(self, data):
        if data["type"] == ServiceType.TerraForm:
            return SidecarTerraformService(**data)

    @validates_schema
    def validate_numbers(self, data):
        if data['type'] != ServiceType.TerraForm:
            raise ValidationError("service '{}' must be of type '{}' only".format(data['name'], ServiceType.TerraForm))

    class Meta:
        strict = True


class SidecarApplicationSchema(Schema):
    default_health_check_ports_to_test = fields.List(fields.Integer())
    healthcheck_timeout = fields.Integer()
    env = fields.Dict(values=fields.Str(), keys=fields.Str())
    dependencies = fields.List(fields.Str())
    instances_count = fields.Integer()
    healthcheck_script = fields.Str()
    has_public_access = fields.Bool()
    outputs = fields.List(fields.Str(), allow_none=True, required=False)

    @post_load
    def make(self, data):
        data["name"] = "yet-to-be-filled"  # remove when json will become json and not dynamic field named
        return SidecarApplication(**data)

    class Meta:
        strict = True


class SidecarConfigurationSchema(Schema):
    environment = fields.Str()
    provider = fields.Str()
    sandbox_id = fields.Str()
    production_id = fields.Str(allow_none=True)
    space_id = fields.Str()
    cloud_external_key = fields.Str()
    apps = fields.Dict(keys=fields.Str(), values=fields.Nested(SidecarApplicationSchema()))
    services = fields.List(fields.Nested(SidecarServiceSchema()))
    messaging = fields.Nested(MessagingConnectionPropertiesSchema(), required=True)
    env_type = fields.Str()
    internet_facing = fields.Bool(required=True)

    @post_load
    def make(self, data):
        for k, v in data["apps"].items():
            v.name = k
        data["apps"] = [v for k, v in data["apps"].items()]

    @validates_schema
    def validate_numbers(self, data):
        if data['apps'] is None or len(data['apps']) == 0:
            raise ValidationError("Cannot have 0 applications")

    class Meta:
        strict = True


class KubernetesSidecarConfigurationSchema(SidecarConfigurationSchema):
    kub_api_address = fields.Str()

    @post_load
    def make(self, data):
        super().make(data)
        data.pop('provider', None)
        return KubernetesSidecarConfiguration(**data)

    class Meta:
        strict = True


class AwsSidecarConfigurationSchema(SidecarConfigurationSchema):
    region_name = fields.Str()
    onboarding_region = fields.Str(allow_none=True)
    data_table_name = fields.Str()
    virtual_network_id = fields.Str()
    infra_stack_name = fields.Str(allow_none=True)

    @post_load
    def make(self, data):
        super().make(data)
        data.pop('provider', None)
        return AwsSidecarConfiguration(**data)

    class Meta:
        strict = True


class AzureSidecarConfigurationSchema(SidecarConfigurationSchema):
    tenant_id = fields.Str(required=False, allow_none=True)
    application_secret = fields.Str(required=False, allow_none=True)
    application_id = fields.Str(required=False, allow_none=True)
    subscription_id = fields.Str()
    management_resource_group = fields.Str()
    vnet_name = fields.Str()
    managed_identity_client_id = fields.Str(required=False, allow_none=True)

    @post_load
    def make(self, data):
        super().make(data)
        data.pop('provider', None)
        return AzureSidecarConfiguration(**data)

    class Meta:
        strict = True
